from dataclasses import InitVar
from typing import Any, Iterable, Optional, Tuple, Union, get_type_hints

from .types import JSON, Vec3D, Vec3Df


def from_json(data: JSON, cls: Optional[type]) -> Any:
    cls_annotations = getattr(cls, "__annotations__", None)
    cls_origin = getattr(cls, "__origin__", None)
    cls_args = getattr(cls, "__args__", None)
    if cls_origin == Union:
        for unioned_cls in cls_args:
            try:
                return from_json(data, unioned_cls)
            except Exception:
                pass
        return data
    if cls_annotations is not None:
        assert cls is not None
        if isinstance(data, dict):
            for name, value in data.items():
                field_type = cls_annotations.get(name)
                data[name] = from_json(value, field_type)
            return cls(**data)
        else:
            return cls(
                *(
                    from_json(field_data, field_type)
                    for field_data, field_type in zip(data, cls_annotations.values())
                )
            )
    elif isinstance(cls_origin, type) and cls_args is not None:
        if issubclass(cls_origin, list):
            list_type = cls_args[0]
            return [from_json(value, list_type) for value in data]
        elif issubclass(cls_origin, tuple):
            if cls_args[-1] == Ellipsis:
                tuple_type = cls_args[0]
                return tuple(from_json(value, tuple_type) for value in data)
            return tuple(
                from_json(value, element_type)
                for value, element_type in zip(data, cls_args)
            )
        elif issubclass(cls_origin, dict):
            key_type = cls_args[0]
            value_type = cls_args[1]
            return {
                from_json(key, key_type): from_json(value, value_type)
                for key, value in data.items()
            }
    else:
        return data


def yield_jsons(
    obj_iter: Iterable, keys: Optional[Iterable[str]] = None
) -> Iterable[Union[JSON, Tuple[str, JSON]]]:
    if keys is None:
        for obj in obj_iter:
            json = to_json(obj)
            if json is not None:
                yield json
    else:
        for key, obj in zip(keys, obj_iter):
            json = to_json(obj)
            if json is not None:
                yield key, json


def to_json(obj: Any) -> JSON:
    cls = type(obj)
    if issubclass(cls, (Vec3D, Vec3Df)):
        return tuple(obj)
    elif hasattr(cls, "__annotations__"):
        annotations = get_type_hints(cls)
        annotations = {k: v for k, v in annotations.items() if v != InitVar}
        return dict(
            yield_jsons(
                (getattr(obj, name) for name in annotations), keys=annotations.keys()
            )
        )
    elif issubclass(cls, list):
        return list(yield_jsons(obj))
    elif issubclass(cls, tuple):
        return tuple(yield_jsons(obj))
    elif issubclass(cls, dict):
        return dict(yield_jsons(obj.values(), keys=obj.keys()))
    else:
        return obj



import asyncio
from typing import List, Tuple

import numpy as np
from sanic import Blueprint, response
from sanic.request import Request
from wkcuber.mag import Mag

from ...utils.json import to_json
from ...utils.types import Vec3D
from ...webknossos.access import AccessRequest, authorized
from ...webknossos.models import BoundingBox as WkBoundingBox
from ...webknossos.models import DataLayer as WKDataLayer
from ...webknossos.models import DataRequest as WKDataRequest
from ...webknossos.models import Histogram

histogram = Blueprint(__name__)

BUCKET_SIZE = 32


@histogram.route(
    "/<organization_name>/<dataset_name>/layers/<layer_name>/histogram", methods=["GET"]
)
@authorized(AccessRequest.read_dataset)
async def histogram_post(
    request: Request, organization_name: str, dataset_name: str, layer_name: str
) -> response.HTTPResponse:
    (backend_name, dataset) = request.app.ctx.repository.get_dataset(
        organization_name, dataset_name
    )
    backend = request.app.ctx.backends[backend_name]

    layer = [i for i in dataset.to_webknossos().dataLayers if i.name == layer_name][0]

    sample_positions = generate_sample_positions(2, layer.boundingBox, BUCKET_SIZE)
    sample_positions, zoom_step = align_positions_with_mag(sample_positions, layer)
    sample_positions_distinct = sorted(set(sample_positions))

    bucket_requests = [
        WKDataRequest(position, zoomStep=zoom_step, cubeSize=BUCKET_SIZE, fourBit=False)
        for position in sample_positions_distinct
    ]

    buckets = await asyncio.gather(
        *(
            backend.read_data(
                dataset,
                layer_name,
                r.zoomStep,
                Vec3D(*r.position),
                Vec3D(r.cubeSize, r.cubeSize, r.cubeSize),
            )
            for r in bucket_requests
        )
    )
    existing_buckets = [
        data for r, data in zip(bucket_requests, buckets) if data is not None
    ]
    data = np.concatenate(existing_buckets) if len(existing_buckets) > 0 else b""
    data = data[np.nonzero(data)]

    if np.issubdtype(data.dtype, np.integer):
        minimum, maximum = np.iinfo(data.dtype).min, np.iinfo(data.dtype).max
        counts, _ = np.histogram(data, bins=np.arange(minimum, maximum))
        histogram = Histogram(
            [int(c) for c in counts], len(data), int(minimum), int(maximum)
        )
    elif np.issubdtype(data.dtype, np.floating):
        minimum, maximum = np.min(data), np.max(data)
        bucket_size = (maximum - minimum) / 255
        bucket_size = 1.0 if np.isclose(bucket_size, 0.0) else bucket_size
        counts, _ = np.histogram(data, bins=np.arange(minimum, maximum, bucket_size))
        histogram = Histogram(
            [int(c) for c in counts], len(data), float(minimum), float(maximum)
        )
    else:
        raise Exception("Histogram for data type {data.dtype} is not supported.")

    return response.json(to_json([histogram]))


def align_positions_with_mag(
    sample_positions: List[Vec3D], layer: WKDataLayer
) -> Tuple[List[Vec3D], int]:
    # For the WKW backend, the bucket requests need to be bucket-aligned in the target mag
    available_mags = sorted([Mag(mag["resolution"]) for mag in layer.wkwResolutions])
    mag = available_mags[0]

    # This is equivalent to `int(log2(mag.as_np().max()))`, but avoids intermediate floats
    zoom_step = int(mag.as_np().max()).bit_length() - 1

    align = Vec3D(*(mag.as_np() * BUCKET_SIZE))
    sample_positions = [
        Vec3D(*((position // align) * align)) for position in sample_positions
    ]
    return sample_positions, zoom_step


def generate_sample_positions(
    iterations: int, bounding_box: WkBoundingBox, resolution_limit: int
) -> List[Vec3D]:
    positions: List[Vec3D] = []
    for exponent in range(1, iterations + 1):
        power = 2 ** exponent
        distance = Vec3D(
            int(bounding_box.width / power),
            int(bounding_box.height / power),
            int(bounding_box.depth / power),
        )
        if (
            distance.x < resolution_limit
            and distance.y < resolution_limit
            and distance.z < resolution_limit
        ):
            return positions
        top_left = bounding_box.topLeft
        positions_for_exponent = []
        for x in range(1, power):
            for y in range(1, power):
                for z in range(1, power):
                    positions_for_exponent.append(top_left + Vec3D(x, y, z) * distance)
        positions.extend(positions_for_exponent)
    return positions



from dataclasses import dataclass
from functools import lru_cache
from typing import Optional, Tuple

import numpy as np
from async_lru import alru_cache
from wkcuber.api.Dataset import WKDataset
from wkcuber.api.Properties.LayerProperties import LayerProperties
from wkcuber.mag import Mag

from ...fast_wkw import DatasetCache, DatasetHandle  # pylint: disable=no-name-in-module
from ...utils.types import Vec3D, Vec3Df
from ...webknossos.models import BoundingBox as WkBoundingBox
from ...webknossos.models import DataLayer as WkDataLayer
from ...webknossos.models import DataSource as WkDataSource
from ...webknossos.models import DataSourceId as WkDataSourceId
from ..backend import DatasetInfo


@dataclass(frozen=True)
class Dataset(DatasetInfo):
    organization_name: str
    dataset_name: str
    dataset_handle: WKDataset
    wkw_cache: DatasetCache

    def to_webknossos(self) -> WkDataSource:
        return WkDataSource(
            WkDataSourceId(self.organization_name, self.dataset_name),
            [
                self.layer_to_webknossos(layer_name, layer)
                for layer_name, layer in self.dataset_handle.properties.data_layers.items()
            ],
            Vec3Df(*self.dataset_handle.properties.scale),
        )

    def layer_to_webknossos(
        self, layer_name: str, layer_properties: LayerProperties
    ) -> WkDataLayer:
        return WkDataLayer(
            layer_name,
            layer_properties._category,
            WkBoundingBox(
                topLeft=Vec3D(*layer_properties._bounding_box["topLeft"]),
                width=layer_properties._bounding_box["width"],
                height=layer_properties._bounding_box["height"],
                depth=layer_properties._bounding_box["depth"],
            ),
            [
                Vec3D(*mag.mag.to_array())
                for mag in layer_properties._wkw_magnifications
            ],
            layer_properties._element_class,
        )

    @lru_cache(maxsize=1000)
    def get_data_handle(
        self, layer_name: str, zoom_step: int
    ) -> Optional[Tuple[DatasetHandle, Mag]]:
        layer = self.dataset_handle.get_layer(layer_name)
        # Finding the right mag for the zoom_step:
        # 2 ** zoomStep = max(mag.x, mag.y, mag.z)
        mag_datasets = [
            mag_dataset
            for mag_dataset in layer.mags.values()
            if (2 ** zoom_step) == Mag(mag_dataset.name).as_np().max()
        ]
        if len(mag_datasets) < 1:
            return None
        mag_dataset = mag_datasets[0]

        data_handle = self.wkw_cache.get_dataset(str(mag_dataset.view.path))
        return (data_handle, Mag(mag_dataset.name))

    @alru_cache(maxsize=2 ** 12, cache_exceptions=False)
    async def read_data(
        self, layer_name: str, zoom_step: int, wk_offset: Vec3D, shape: Vec3D
    ) -> Optional[np.ndarray]:
        assert shape == Vec3D(
            32, 32, 32
        ), "Only buckets of 32 edge length are supported"
        assert shape % 32 == Vec3D(0, 0, 0), "Only 32-aligned buckets are supported"
        data_handle_opt = self.get_data_handle(layer_name, zoom_step)
        if data_handle_opt is None:
            return None
        data_handle, mag = data_handle_opt
        offset = (
            np.array([wk_offset.x, wk_offset.y, wk_offset.z]) / mag.as_np()
        ).astype(np.uint32)
        block = await data_handle.read_block(tuple(offset))
        return np.frombuffer(block.buf, dtype=np.dtype(block.dtype)).reshape(
            block.shape, order="F"
        )

    def clear_cache(self) -> None:
        self.read_data.cache_clear()  # pylint: disable=no-member
        self.get_data_handle.cache_clear()  # pylint: disable=no-member
        self.wkw_cache.clear_cache_prefix(str(self.dataset_handle.path))



import asyncio
import json
import os
from typing import TYPE_CHECKING, Any, Iterable, Tuple, cast

from sanic import Blueprint, response
from sanic.request import Request

from ...utils.types import JSON
from ...webknossos.access import AccessRequest, authorized

if TYPE_CHECKING:
    from ...__main__ import Server

upload = Blueprint(__name__)


def iterate_datasets(datasets: JSON) -> Iterable[Tuple[Any, str, str, str]]:
    for backend_name, backend_vals in datasets.items():
        for organization_name, organization_vals in backend_vals.items():
            for dataset_name, dataset_config in organization_vals.items():
                yield dataset_config, backend_name, organization_name, dataset_name


@upload.route("", methods=["POST"])
@authorized(AccessRequest.administrate_datasets)
async def add_dataset(request: Request) -> response.HTTPResponse:
    app = cast("Server", request.app)
    await asyncio.gather(
        *(
            app.add_dataset(*dataset_args)
            for dataset_args in iterate_datasets(request.json)
        )
    )

    ds_path = app.config["datasets_path"]
    os.makedirs(os.path.dirname(ds_path), exist_ok=True)
    if not os.path.isfile(ds_path):
        with open(ds_path, "w") as datasets_file:
            json.dump({}, datasets_file)
    with open(ds_path, "r+") as datasets_file:
        datasets = json.load(datasets_file)
        datasets_file.seek(0)
        for (
            dataset_config,
            backend_name,
            organization_name,
            dataset_name,
        ) in iterate_datasets(request.json):
            datasets.setdefault(backend_name, {}).setdefault(organization_name, {})[
                dataset_name
            ] = dataset_config
        json.dump(datasets, datasets_file, indent=4, sort_keys=True)
        # remove rest of file, as it might be shorter
        datasets_file.truncate()

    return response.text("")



from sanic import Blueprint, response
from sanic.request import Request

from ...utils.types import Vec3D
from ...webknossos.access import AccessRequest, authorized

meshes = Blueprint(__name__)


@meshes.route(
    "/<organization_name>/<dataset_name>/layers/<layer_name>/meshes", methods=["GET"]
)
@authorized(AccessRequest.read_dataset)
async def get_meshes(
    request: Request, organization_name: str, dataset_name: str, layer_name: str
) -> response.HTTPResponse:
    (backend_name, dataset) = request.app.ctx.repository.get_dataset(
        organization_name, dataset_name
    )
    backend = request.app.ctx.backends[backend_name]

    try:
        return response.json(await backend.get_meshes(dataset, layer_name))
    except NotImplementedError:
        return response.empty(status=501)


@meshes.route(
    "/<organization_name>/<dataset_name>/layers/<layer_name>/meshes/chunks",
    methods=["POST"],
)
@authorized(AccessRequest.read_dataset)
async def get_mesh_chunks(
    request: Request, organization_name: str, dataset_name: str, layer_name: str
) -> response.HTTPResponse:
    request_body = request.json

    (backend_name, dataset) = request.app.ctx.repository.get_dataset(
        organization_name, dataset_name
    )
    backend = request.app.ctx.backends[backend_name]

    segment_id = request_body["segmentId"]
    mesh_name = request_body["meshFile"]
    try:
        chunks = await backend.get_chunks_for_mesh(
            dataset, layer_name, mesh_name, segment_id
        )
        if chunks is None:
            return response.empty(status=404)
        return response.json(chunks)
    except NotImplementedError:
        return response.empty(status=501)


@meshes.route(
    "/<organization_name>/<dataset_name>/layers/<layer_name>/meshes/chunks/data",
    methods=["POST"],
)
@authorized(AccessRequest.read_dataset)
async def get_mesh_chunk_data(
    request: Request, organization_name: str, dataset_name: str, layer_name: str
) -> response.HTTPResponse:
    request_body = request.json

    (backend_name, dataset) = request.app.ctx.repository.get_dataset(
        organization_name, dataset_name
    )
    backend = request.app.ctx.backends[backend_name]
    segment_id = request_body["segmentId"]
    mesh_name = request_body["meshFile"]
    position = Vec3D(*request_body["position"])

    try:
        chunk_data = await backend.get_chunk_data_for_mesh(
            dataset, layer_name, mesh_name, segment_id, position
        )
        if chunk_data is None:
            return response.empty(status=404)

        return response.raw(chunk_data)
    except NotImplementedError:
        return response.empty(status=501)



import time
from functools import wraps
from typing import Any, Awaitable, Callable, TypeVar, cast

from async_lru import alru_cache

T = TypeVar("T", bound=Callable[..., Awaitable[Any]])


def atlru_cache(
    seconds_to_use: int, maxsize: int = 512, typed: bool = False
) -> Callable[[T], T]:
    def decorator(f: T) -> T:
        @wraps(f)
        @alru_cache(maxsize=maxsize, typed=typed, cache_exceptions=False)
        async def time_f(*args: Any, **kwargs: Any) -> Any:
            return (time.monotonic(), await f(*args, **kwargs))

        @wraps(time_f)
        async def cache_decorated_function(*args: Any, **kwargs: Any) -> Any:
            now = time.monotonic()
            inserted, result = await time_f(*args, **kwargs)
            if now - inserted > seconds_to_use:
                time_f.invalidate(*args, **kwargs)  # type: ignore
                new_inserted, result = await time_f(*args, **kwargs)
                assert new_inserted >= now
            return result

        # cast necessary, see
        # https://github.com/python/mypy/issues/1927
        return cast(T, cache_decorated_function)

    return decorator



import asyncio
import json
from typing import List

import numpy as np
from sanic import Blueprint, response
from sanic.request import Request

from ...utils.json import from_json
from ...utils.types import Vec3D
from ...webknossos.access import AccessRequest, authorized
from ...webknossos.models import DataRequest as WKDataRequest

read_data = Blueprint(__name__)


def convert_data(to_four_bit: bool, data: np.ndarray) -> np.ndarray:
    data = data.flatten(order="F")
    if to_four_bit:
        assert data.dtype == np.dtype("uint8")
        odds = data[::2]
        evens = data[1::2]
        return ((odds >> 4) << 4) | (evens >> 4)
    elif data.dtype == np.dtype("uint64"):
        return (data & np.uint64(0xFFFFFFFF)).astype("uint32")
    else:
        return data


@read_data.route(
    "/<organization_name>/<dataset_name>/layers/<layer_name>/data", methods=["POST"]
)
@authorized(AccessRequest.read_dataset)
async def get_data_post(
    request: Request, organization_name: str, dataset_name: str, layer_name: str
) -> response.HTTPResponse:
    (backend_name, dataset) = request.app.ctx.repository.get_dataset(
        organization_name, dataset_name
    )
    backend = request.app.ctx.backends[backend_name]

    bucket_requests = from_json(request.json, List[WKDataRequest])

    buckets = await asyncio.gather(
        *(
            backend.read_data(
                dataset,
                layer_name,
                r.zoomStep,
                Vec3D(*r.position),
                Vec3D(r.cubeSize, r.cubeSize, r.cubeSize),
            )
            for r in bucket_requests
        )
    )
    missing_buckets = [index for index, data in enumerate(buckets) if data is None]
    existing_buckets = [
        convert_data(r.fourBit, data)
        for r, data in zip(bucket_requests, buckets)
        if data is not None
    ]
    data = (
        np.concatenate(existing_buckets).tobytes() if len(existing_buckets) > 0 else b""
    )

    headers = {
        "Access-Control-Expose-Headers": "MISSING-BUCKETS",
        "MISSING-BUCKETS": json.dumps(missing_buckets),
    }
    return response.raw(data, headers=headers)



from sanic import Blueprint, response
from sanic.request import Request

from ...utils.json import to_json
from ...webknossos.access import AccessRequest, authorized

datasource_properties = Blueprint(__name__)


@datasource_properties.route("/<organization_name>/<dataset_name>", methods=["GET"])
@authorized(AccessRequest.read_dataset)
async def datasource_properties_get(
    request: Request, organization_name: str, dataset_name: str
) -> response.HTTPResponse:
    (_, dataset) = request.app.ctx.repository.get_dataset(
        organization_name, dataset_name
    )
    return response.json(
        {"dataSource": to_json(dataset.to_webknossos()), "messages": []}
    )


@datasource_properties.route("/<organization_name>/<dataset_name>", methods=["POST"])
async def datasource_properties_post(
    request: Request, organization_name: str, dataset_name: str
) -> response.HTTPResponse:
    return response.text(
        "Not Implemented: webknossos-connect does not support editing the datasource-properties. Try refreshing the dataset instead.",
        status=501,
    )



from dataclasses import dataclass, replace
from io import BytesIO
from typing import Any, List, Tuple

import DracoPy
import numpy as np
from stl import Mesh

from ...utils.types import Vec3Df
from .sharding import ShardingInfo


@dataclass(frozen=True)
class MeshfileFragment:
    position: Vec3Df
    byte_offset: int
    byte_size: int


@dataclass(frozen=True)
class MeshfileLod:
    scale: int
    vertex_offset: Vec3Df
    chunk_shape: Vec3Df
    fragments: List[MeshfileFragment]


@dataclass(frozen=True)
class Meshfile:
    chunk_shape: Vec3Df
    grid_origin: Vec3Df
    lods: List[MeshfileLod]

    def fragment_offset_and_shape(
        self, lod: int, fragment: int
    ) -> Tuple[Vec3Df, Vec3Df, Vec3Df]:
        return (
            self.grid_origin
            + self.lods[lod].fragments[fragment].position
            * self.chunk_shape
            * self.lods[lod].scale,
            self.grid_origin
            + self.lods[lod].fragments[fragment].position
            * self.chunk_shape
            * self.lods[lod].scale
            + self.chunk_shape * self.lods[lod].scale,
            self.chunk_shape * self.lods[lod].scale,
        )

    def decode_data(self, fragment: MeshfileFragment, buf: bytes) -> bytes:
        mesh_obj = DracoPy.decode_buffer_to_mesh(buf)
        vertices = np.array(mesh_obj.points, dtype="f4").reshape((-1, 3))
        faces = np.array(mesh_obj.faces, dtype="<u4").reshape((-1, 3))
        stl_mesh = Mesh(np.zeros(faces.shape[0], dtype=Mesh.dtype))
        for i, f in enumerate(faces):
            for j in range(3):
                stl_mesh.vectors[i][j] = fragment.position.as_np() + vertices[f[j], :]
        out_buf = BytesIO()
        stl_mesh.save("mesh.stl", out_buf, update_normals=True)
        return out_buf.getvalue()


@dataclass(frozen=True)
class MeshInfo:
    """
    Stores the representation of the meshes of a segmentation layer.
    The format is defined at https://github.com/google/neuroglancer/blob/master/src/neuroglancer/datasource/precomputed/meshes.md.
    """

    sharding: ShardingInfo
    _transform: Tuple[float, ...]
    vertex_quantization_bits: int
    lod_scale_multiplier: int = 1

    @staticmethod
    def parse(info_json: Any) -> "MeshInfo":
        assert info_json["@type"] == "neuroglancer_multilod_draco"
        return MeshInfo(
            sharding=ShardingInfo.parse(info_json["sharding"]),
            _transform=tuple(info_json["transform"]),
            vertex_quantization_bits=info_json["vertex_quantization_bits"],
            lod_scale_multiplier=info_json["lod_scale_multiplier"],
        )

    @property
    def transform(self) -> np.ndarray:
        return np.array(self._transform, dtype="f4").reshape((3, 4))

    def parse_meshfile(self, buf: bytes, byte_offset: int) -> Meshfile:
        p = 0
        chunk_shape = np.frombuffer(buf[p : p + 12], dtype="f4").copy()
        p += 12
        grid_origin = np.frombuffer(buf[p : p + 12], dtype="f4").copy()
        p += 12
        num_lods = int(np.frombuffer(buf[p : p + 4], dtype="<u4").copy())
        p += 4
        lod_scales = np.frombuffer(buf[p : (p + num_lods * 4)], dtype="f4").copy()
        p += num_lods * 4
        vertex_offsets = (
            np.frombuffer(buf[p : (p + num_lods * 3 * 4)], dtype="f4")
            .reshape((-1, 3))
            .copy()
        )
        p += num_lods * 3 * 4
        num_fragments_per_lod = np.frombuffer(
            buf[p : (p + num_lods * 4)], dtype="<u4"
        ).copy()
        p += num_lods * 4
        lods = []
        frag_byte_offset = 0
        for lod in range(num_lods):
            if num_fragments_per_lod[lod] > 0:
                fragment_positions = (
                    np.frombuffer(
                        buf[p : (p + int(num_fragments_per_lod[lod]) * 3 * 4)],
                        dtype="<u4",
                    )
                    .reshape((3, -1))
                    .T.copy()
                )
                p += int(num_fragments_per_lod[lod]) * 3 * 4
                fragment_byte_sizes = np.frombuffer(
                    buf[p : (p + int(num_fragments_per_lod[lod]) * 4)], dtype="<u4"
                ).copy()
                fragment_byte_offsets = np.zeros(
                    (num_fragments_per_lod[lod],), dtype="<u4"
                )
                fragment_byte_offsets[0] = frag_byte_offset
                fragment_byte_offsets[1:] = frag_byte_offset + np.cumsum(
                    fragment_byte_sizes[:-1]
                )
                frag_byte_offset += np.sum(fragment_byte_sizes)
                p += int(num_fragments_per_lod[lod]) * 4

                fragments = []
                for i in range(num_fragments_per_lod[lod]):
                    global_position = (
                        grid_origin
                        + fragment_positions[i] * chunk_shape * lod_scales[lod]
                    )
                    fragments.append(
                        MeshfileFragment(
                            position=Vec3Df(
                                *np.matmul(global_position, self.transform)[0:3]
                            ),
                            byte_offset=fragment_byte_offsets[i],
                            byte_size=int(fragment_byte_sizes[i]),
                        )
                    )

                lods.append(
                    MeshfileLod(
                        scale=int(lod_scales[lod]),
                        vertex_offset=Vec3Df(*vertex_offsets[lod]),
                        chunk_shape=Vec3Df(*chunk_shape * lod_scales[lod]),
                        fragments=fragments,
                    )
                )

        return Meshfile(
            chunk_shape=Vec3Df(*chunk_shape),
            grid_origin=Vec3Df(*grid_origin),
            lods=[
                replace(
                    lod,
                    fragments=[
                        replace(
                            frag,
                            byte_offset=int(
                                byte_offset - frag_byte_offset + frag.byte_offset
                            ),
                        )
                        for frag in lod.fragments
                    ],
                )
                for lod in lods
            ],
        )



from collections import OrderedDict

# taken from https://xkcd.com/color/rgb/

colors = OrderedDict(
    [
        ("cloudy blue", "acc2d9"),
        ("dark pastel green", "56ae57"),
        ("dust", "b2996e"),
        ("electric lime", "a8ff04"),
        ("fresh green", "69d84f"),
        ("light eggplant", "894585"),
        ("nasty green", "70b23f"),
        ("really light blue", "d4ffff"),
        ("tea", "65ab7c"),
        ("warm purple", "952e8f"),
        ("yellowish tan", "fcfc81"),
        ("cement", "a5a391"),
        ("dark grass green", "388004"),
        ("dusty teal", "4c9085"),
        ("grey teal", "5e9b8a"),
        ("macaroni and cheese", "efb435"),
        ("pinkish tan", "d99b82"),
        ("spruce", "0a5f38"),
        ("strong blue", "0c06f7"),
        ("toxic green", "61de2a"),
        ("windows blue", "3778bf"),
        ("blue blue", "2242c7"),
        ("blue with a hint of purple", "533cc6"),
        ("booger", "9bb53c"),
        ("bright sea green", "05ffa6"),
        ("dark green blue", "1f6357"),
        ("deep turquoise", "017374"),
        ("green teal", "0cb577"),
        ("strong pink", "ff0789"),
        ("bland", "afa88b"),
        ("deep aqua", "08787f"),
        ("lavender pink", "dd85d7"),
        ("light moss green", "a6c875"),
        ("light seafoam green", "a7ffb5"),
        ("olive yellow", "c2b709"),
        ("pig pink", "e78ea5"),
        ("deep lilac", "966ebd"),
        ("desert", "ccad60"),
        ("dusty lavender", "ac86a8"),
        ("purpley grey", "947e94"),
        ("purply", "983fb2"),
        ("candy pink", "ff63e9"),
        ("light pastel green", "b2fba5"),
        ("boring green", "63b365"),
        ("kiwi green", "8ee53f"),
        ("light grey green", "b7e1a1"),
        ("orange pink", "ff6f52"),
        ("tea green", "bdf8a3"),
        ("very light brown", "d3b683"),
        ("egg shell", "fffcc4"),
        ("eggplant purple", "430541"),
        ("powder pink", "ffb2d0"),
        ("reddish grey", "997570"),
        ("baby shit brown", "ad900d"),
        ("liliac", "c48efd"),
        ("stormy blue", "507b9c"),
        ("ugly brown", "7d7103"),
        ("custard", "fffd78"),
        ("darkish pink", "da467d"),
        ("deep brown", "410200"),
        ("greenish beige", "c9d179"),
        ("manilla", "fffa86"),
        ("off blue", "5684ae"),
        ("battleship grey", "6b7c85"),
        ("browny green", "6f6c0a"),
        ("bruise", "7e4071"),
        ("kelley green", "009337"),
        ("sickly yellow", "d0e429"),
        ("sunny yellow", "fff917"),
        ("azul", "1d5dec"),
        ("darkgreen", "054907"),
        ("green/yellow", "b5ce08"),
        ("lichen", "8fb67b"),
        ("light light green", "c8ffb0"),
        ("pale gold", "fdde6c"),
        ("sun yellow", "ffdf22"),
        ("tan green", "a9be70"),
        ("burple", "6832e3"),
        ("butterscotch", "fdb147"),
        ("toupe", "c7ac7d"),
        ("dark cream", "fff39a"),
        ("indian red", "850e04"),
        ("light lavendar", "efc0fe"),
        ("poison green", "40fd14"),
        ("baby puke green", "b6c406"),
        ("bright yellow green", "9dff00"),
        ("charcoal grey", "3c4142"),
        ("squash", "f2ab15"),
        ("cinnamon", "ac4f06"),
        ("light pea green", "c4fe82"),
        ("radioactive green", "2cfa1f"),
        ("raw sienna", "9a6200"),
        ("baby purple", "ca9bf7"),
        ("cocoa", "875f42"),
        ("light royal blue", "3a2efe"),
        ("orangeish", "fd8d49"),
        ("rust brown", "8b3103"),
        ("sand brown", "cba560"),
        ("swamp", "698339"),
        ("tealish green", "0cdc73"),
        ("burnt siena", "b75203"),
        ("camo", "7f8f4e"),
        ("dusk blue", "26538d"),
        ("fern", "63a950"),
        ("old rose", "c87f89"),
        ("pale light green", "b1fc99"),
        ("peachy pink", "ff9a8a"),
        ("rosy pink", "f6688e"),
        ("light bluish green", "76fda8"),
        ("light bright green", "53fe5c"),
        ("light neon green", "4efd54"),
        ("light seafoam", "a0febf"),
        ("tiffany blue", "7bf2da"),
        ("washed out green", "bcf5a6"),
        ("browny orange", "ca6b02"),
        ("nice blue", "107ab0"),
        ("sapphire", "2138ab"),
        ("greyish teal", "719f91"),
        ("orangey yellow", "fdb915"),
        ("parchment", "fefcaf"),
        ("straw", "fcf679"),
        ("very dark brown", "1d0200"),
        ("terracota", "cb6843"),
        ("ugly blue", "31668a"),
        ("clear blue", "247afd"),
        ("creme", "ffffb6"),
        ("foam green", "90fda9"),
        ("grey/green", "86a17d"),
        ("light gold", "fddc5c"),
        ("seafoam blue", "78d1b6"),
        ("topaz", "13bbaf"),
        ("violet pink", "fb5ffc"),
        ("wintergreen", "20f986"),
        ("yellow tan", "ffe36e"),
        ("dark fuchsia", "9d0759"),
        ("indigo blue", "3a18b1"),
        ("light yellowish green", "c2ff89"),
        ("pale magenta", "d767ad"),
        ("rich purple", "720058"),
        ("sunflower yellow", "ffda03"),
        ("green/blue", "01c08d"),
        ("leather", "ac7434"),
        ("racing green", "014600"),
        ("vivid purple", "9900fa"),
        ("dark royal blue", "02066f"),
        ("hazel", "8e7618"),
        ("muted pink", "d1768f"),
        ("booger green", "96b403"),
        ("canary", "fdff63"),
        ("cool grey", "95a3a6"),
        ("dark taupe", "7f684e"),
        ("darkish purple", "751973"),
        ("true green", "089404"),
        ("coral pink", "ff6163"),
        ("dark sage", "598556"),
        ("dark slate blue", "214761"),
        ("flat blue", "3c73a8"),
        ("mushroom", "ba9e88"),
        ("rich blue", "021bf9"),
        ("dirty purple", "734a65"),
        ("greenblue", "23c48b"),
        ("icky green", "8fae22"),
        ("light khaki", "e6f2a2"),
        ("warm blue", "4b57db"),
        ("dark hot pink", "d90166"),
        ("deep sea blue", "015482"),
        ("carmine", "9d0216"),
        ("dark yellow green", "728f02"),
        ("pale peach", "ffe5ad"),
        ("plum purple", "4e0550"),
        ("golden rod", "f9bc08"),
        ("neon red", "ff073a"),
        ("old pink", "c77986"),
        ("very pale blue", "d6fffe"),
        ("blood orange", "fe4b03"),
        ("grapefruit", "fd5956"),
        ("sand yellow", "fce166"),
        ("clay brown", "b2713d"),
        ("dark blue grey", "1f3b4d"),
        ("flat green", "699d4c"),
        ("light green blue", "56fca2"),
        ("warm pink", "fb5581"),
        ("dodger blue", "3e82fc"),
        ("gross green", "a0bf16"),
        ("ice", "d6fffa"),
        ("metallic blue", "4f738e"),
        ("pale salmon", "ffb19a"),
        ("sap green", "5c8b15"),
        ("algae", "54ac68"),
        ("bluey grey", "89a0b0"),
        ("greeny grey", "7ea07a"),
        ("highlighter green", "1bfc06"),
        ("light light blue", "cafffb"),
        ("light mint", "b6ffbb"),
        ("raw umber", "a75e09"),
        ("vivid blue", "152eff"),
        ("deep lavender", "8d5eb7"),
        ("dull teal", "5f9e8f"),
        ("light greenish blue", "63f7b4"),
        ("mud green", "606602"),
        ("pinky", "fc86aa"),
        ("red wine", "8c0034"),
        ("shit green", "758000"),
        ("tan brown", "ab7e4c"),
        ("darkblue", "030764"),
        ("rosa", "fe86a4"),
        ("lipstick", "d5174e"),
        ("pale mauve", "fed0fc"),
        ("claret", "680018"),
        ("dandelion", "fedf08"),
        ("orangered", "fe420f"),
        ("poop green", "6f7c00"),
        ("ruby", "ca0147"),
        ("dark", "1b2431"),
        ("greenish turquoise", "00fbb0"),
        ("pastel red", "db5856"),
        ("piss yellow", "ddd618"),
        ("bright cyan", "41fdfe"),
        ("dark coral", "cf524e"),
        ("algae green", "21c36f"),
        ("darkish red", "a90308"),
        ("reddy brown", "6e1005"),
        ("blush pink", "fe828c"),
        ("camouflage green", "4b6113"),
        ("lawn green", "4da409"),
        ("putty", "beae8a"),
        ("vibrant blue", "0339f8"),
        ("dark sand", "a88f59"),
        ("purple/blue", "5d21d0"),
        ("saffron", "feb209"),
        ("twilight", "4e518b"),
        ("warm brown", "964e02"),
        ("bluegrey", "85a3b2"),
        ("bubble gum pink", "ff69af"),
        ("duck egg blue", "c3fbf4"),
        ("greenish cyan", "2afeb7"),
        ("petrol", "005f6a"),
        ("royal", "0c1793"),
        ("butter", "ffff81"),
        ("dusty orange", "f0833a"),
        ("off yellow", "f1f33f"),
        ("pale olive green", "b1d27b"),
        ("orangish", "fc824a"),
        ("leaf", "71aa34"),
        ("light blue grey", "b7c9e2"),
        ("dried blood", "4b0101"),
        ("lightish purple", "a552e6"),
        ("rusty red", "af2f0d"),
        ("lavender blue", "8b88f8"),
        ("light grass green", "9af764"),
        ("light mint green", "a6fbb2"),
        ("sunflower", "ffc512"),
        ("velvet", "750851"),
        ("brick orange", "c14a09"),
        ("lightish red", "fe2f4a"),
        ("pure blue", "0203e2"),
        ("twilight blue", "0a437a"),
        ("violet red", "a50055"),
        ("yellowy brown", "ae8b0c"),
        ("carnation", "fd798f"),
        ("muddy yellow", "bfac05"),
        ("dark seafoam green", "3eaf76"),
        ("deep rose", "c74767"),
        ("dusty red", "b9484e"),
        ("grey/blue", "647d8e"),
        ("lemon lime", "bffe28"),
        ("purple/pink", "d725de"),
        ("brown yellow", "b29705"),
        ("purple brown", "673a3f"),
        ("wisteria", "a87dc2"),
        ("banana yellow", "fafe4b"),
        ("lipstick red", "c0022f"),
        ("water blue", "0e87cc"),
        ("brown grey", "8d8468"),
        ("vibrant purple", "ad03de"),
        ("baby green", "8cff9e"),
        ("barf green", "94ac02"),
        ("eggshell blue", "c4fff7"),
        ("sandy yellow", "fdee73"),
        ("cool green", "33b864"),
        ("pale", "fff9d0"),
        ("blue/grey", "758da3"),
        ("hot magenta", "f504c9"),
        ("greyblue", "77a1b5"),
        ("purpley", "8756e4"),
        ("baby shit green", "889717"),
        ("brownish pink", "c27e79"),
        ("dark aquamarine", "017371"),
        ("diarrhea", "9f8303"),
        ("light mustard", "f7d560"),
        ("pale sky blue", "bdf6fe"),
        ("turtle green", "75b84f"),
        ("bright olive", "9cbb04"),
        ("dark grey blue", "29465b"),
        ("greeny brown", "696006"),
        ("lemon green", "adf802"),
        ("light periwinkle", "c1c6fc"),
        ("seaweed green", "35ad6b"),
        ("sunshine yellow", "fffd37"),
        ("ugly purple", "a442a0"),
        ("medium pink", "f36196"),
        ("puke brown", "947706"),
        ("very light pink", "fff4f2"),
        ("viridian", "1e9167"),
        ("bile", "b5c306"),
        ("faded yellow", "feff7f"),
        ("very pale green", "cffdbc"),
        ("vibrant green", "0add08"),
        ("bright lime", "87fd05"),
        ("spearmint", "1ef876"),
        ("light aquamarine", "7bfdc7"),
        ("light sage", "bcecac"),
        ("yellowgreen", "bbf90f"),
        ("baby poo", "ab9004"),
        ("dark seafoam", "1fb57a"),
        ("deep teal", "00555a"),
        ("heather", "a484ac"),
        ("rust orange", "c45508"),
        ("dirty blue", "3f829d"),
        ("fern green", "548d44"),
        ("bright lilac", "c95efb"),
        ("weird green", "3ae57f"),
        ("peacock blue", "016795"),
        ("avocado green", "87a922"),
        ("faded orange", "f0944d"),
        ("grape purple", "5d1451"),
        ("hot green", "25ff29"),
        ("lime yellow", "d0fe1d"),
        ("mango", "ffa62b"),
        ("shamrock", "01b44c"),
        ("bubblegum", "ff6cb5"),
        ("purplish brown", "6b4247"),
        ("vomit yellow", "c7c10c"),
        ("pale cyan", "b7fffa"),
        ("key lime", "aeff6e"),
        ("tomato red", "ec2d01"),
        ("lightgreen", "76ff7b"),
        ("merlot", "730039"),
        ("night blue", "040348"),
        ("purpleish pink", "df4ec8"),
        ("apple", "6ecb3c"),
        ("baby poop green", "8f9805"),
        ("green apple", "5edc1f"),
        ("heliotrope", "d94ff5"),
        ("yellow/green", "c8fd3d"),
        ("almost black", "070d0d"),
        ("cool blue", "4984b8"),
        ("leafy green", "51b73b"),
        ("mustard brown", "ac7e04"),
        ("dusk", "4e5481"),
        ("dull brown", "876e4b"),
        ("frog green", "58bc08"),
        ("vivid green", "2fef10"),
        ("bright light green", "2dfe54"),
        ("fluro green", "0aff02"),
        ("kiwi", "9cef43"),
        ("seaweed", "18d17b"),
        ("navy green", "35530a"),
        ("ultramarine blue", "1805db"),
        ("iris", "6258c4"),
        ("pastel orange", "ff964f"),
        ("yellowish orange", "ffab0f"),
        ("perrywinkle", "8f8ce7"),
        ("tealish", "24bca8"),
        ("dark plum", "3f012c"),
        ("pear", "cbf85f"),
        ("pinkish orange", "ff724c"),
        ("midnight purple", "280137"),
        ("light urple", "b36ff6"),
        ("dark mint", "48c072"),
        ("greenish tan", "bccb7a"),
        ("light burgundy", "a8415b"),
        ("turquoise blue", "06b1c4"),
        ("ugly pink", "cd7584"),
        ("sandy", "f1da7a"),
        ("electric pink", "ff0490"),
        ("muted purple", "805b87"),
        ("mid green", "50a747"),
        ("greyish", "a8a495"),
        ("neon yellow", "cfff04"),
        ("banana", "ffff7e"),
        ("carnation pink", "ff7fa7"),
        ("tomato", "ef4026"),
        ("sea", "3c9992"),
        ("muddy brown", "886806"),
        ("turquoise green", "04f489"),
        ("buff", "fef69e"),
        ("fawn", "cfaf7b"),
        ("muted blue", "3b719f"),
        ("pale rose", "fdc1c5"),
        ("dark mint green", "20c073"),
        ("amethyst", "9b5fc0"),
        ("blue/green", "0f9b8e"),
        ("chestnut", "742802"),
        ("sick green", "9db92c"),
        ("pea", "a4bf20"),
        ("rusty orange", "cd5909"),
        ("stone", "ada587"),
        ("rose red", "be013c"),
        ("pale aqua", "b8ffeb"),
        ("deep orange", "dc4d01"),
        ("earth", "a2653e"),
        ("mossy green", "638b27"),
        ("grassy green", "419c03"),
        ("pale lime green", "b1ff65"),
        ("light grey blue", "9dbcd4"),
        ("pale grey", "fdfdfe"),
        ("asparagus", "77ab56"),
        ("blueberry", "464196"),
        ("purple red", "990147"),
        ("pale lime", "befd73"),
        ("greenish teal", "32bf84"),
        ("caramel", "af6f09"),
        ("deep magenta", "a0025c"),
        ("light peach", "ffd8b1"),
        ("milk chocolate", "7f4e1e"),
        ("ocher", "bf9b0c"),
        ("off green", "6ba353"),
        ("purply pink", "f075e6"),
        ("lightblue", "7bc8f6"),
        ("dusky blue", "475f94"),
        ("golden", "f5bf03"),
        ("light beige", "fffeb6"),
        ("butter yellow", "fffd74"),
        ("dusky purple", "895b7b"),
        ("french blue", "436bad"),
        ("ugly yellow", "d0c101"),
        ("greeny yellow", "c6f808"),
        ("orangish red", "f43605"),
        ("shamrock green", "02c14d"),
        ("orangish brown", "b25f03"),
        ("tree green", "2a7e19"),
        ("deep violet", "490648"),
        ("gunmetal", "536267"),
        ("blue/purple", "5a06ef"),
        ("cherry", "cf0234"),
        ("sandy brown", "c4a661"),
        ("warm grey", "978a84"),
        ("dark indigo", "1f0954"),
        ("midnight", "03012d"),
        ("bluey green", "2bb179"),
        ("grey pink", "c3909b"),
        ("soft purple", "a66fb5"),
        ("blood", "770001"),
        ("brown red", "922b05"),
        ("medium grey", "7d7f7c"),
        ("berry", "990f4b"),
        ("poo", "8f7303"),
        ("purpley pink", "c83cb9"),
        ("light salmon", "fea993"),
        ("snot", "acbb0d"),
        ("easter purple", "c071fe"),
        ("light yellow green", "ccfd7f"),
        ("dark navy blue", "00022e"),
        ("drab", "828344"),
        ("light rose", "ffc5cb"),
        ("rouge", "ab1239"),
        ("purplish red", "b0054b"),
        ("slime green", "99cc04"),
        ("baby poop", "937c00"),
        ("irish green", "019529"),
        ("pink/purple", "ef1de7"),
        ("dark navy", "000435"),
        ("greeny blue", "42b395"),
        ("light plum", "9d5783"),
        ("pinkish grey", "c8aca9"),
        ("dirty orange", "c87606"),
        ("rust red", "aa2704"),
        ("pale lilac", "e4cbff"),
        ("orangey red", "fa4224"),
        ("primary blue", "0804f9"),
        ("kermit green", "5cb200"),
        ("brownish purple", "76424e"),
        ("murky green", "6c7a0e"),
        ("wheat", "fbdd7e"),
        ("very dark purple", "2a0134"),
        ("bottle green", "044a05"),
        ("watermelon", "fd4659"),
        ("deep sky blue", "0d75f8"),
        ("fire engine red", "fe0002"),
        ("yellow ochre", "cb9d06"),
        ("pumpkin orange", "fb7d07"),
        ("pale olive", "b9cc81"),
        ("light lilac", "edc8ff"),
        ("lightish green", "61e160"),
        ("carolina blue", "8ab8fe"),
        ("mulberry", "920a4e"),
        ("shocking pink", "fe02a2"),
        ("auburn", "9a3001"),
        ("bright lime green", "65fe08"),
        ("celadon", "befdb7"),
        ("pinkish brown", "b17261"),
        ("poo brown", "885f01"),
        ("bright sky blue", "02ccfe"),
        ("celery", "c1fd95"),
        ("dirt brown", "836539"),
        ("strawberry", "fb2943"),
        ("dark lime", "84b701"),
        ("copper", "b66325"),
        ("medium brown", "7f5112"),
        ("muted green", "5fa052"),
        ("robin's egg", "6dedfd"),
        ("bright aqua", "0bf9ea"),
        ("bright lavender", "c760ff"),
        ("ivory", "ffffcb"),
        ("very light purple", "f6cefc"),
        ("light navy", "155084"),
        ("pink red", "f5054f"),
        ("olive brown", "645403"),
        ("poop brown", "7a5901"),
        ("mustard green", "a8b504"),
        ("ocean green", "3d9973"),
        ("very dark blue", "000133"),
        ("dusty green", "76a973"),
        ("light navy blue", "2e5a88"),
        ("minty green", "0bf77d"),
        ("adobe", "bd6c48"),
        ("barney", "ac1db8"),
        ("jade green", "2baf6a"),
        ("bright light blue", "26f7fd"),
        ("light lime", "aefd6c"),
        ("dark khaki", "9b8f55"),
        ("orange yellow", "ffad01"),
        ("ocre", "c69c04"),
        ("maize", "f4d054"),
        ("faded pink", "de9dac"),
        ("british racing green", "05480d"),
        ("sandstone", "c9ae74"),
        ("mud brown", "60460f"),
        ("light sea green", "98f6b0"),
        ("robin egg blue", "8af1fe"),
        ("aqua marine", "2ee8bb"),
        ("dark sea green", "11875d"),
        ("soft pink", "fdb0c0"),
        ("orangey brown", "b16002"),
        ("cherry red", "f7022a"),
        ("burnt yellow", "d5ab09"),
        ("brownish grey", "86775f"),
        ("camel", "c69f59"),
        ("purplish grey", "7a687f"),
        ("marine", "042e60"),
        ("greyish pink", "c88d94"),
        ("pale turquoise", "a5fbd5"),
        ("pastel yellow", "fffe71"),
        ("bluey purple", "6241c7"),
        ("canary yellow", "fffe40"),
        ("faded red", "d3494e"),
        ("sepia", "985e2b"),
        ("coffee", "a6814c"),
        ("bright magenta", "ff08e8"),
        ("mocha", "9d7651"),
        ("ecru", "feffca"),
        ("purpleish", "98568d"),
        ("cranberry", "9e003a"),
        ("darkish green", "287c37"),
        ("brown orange", "b96902"),
        ("dusky rose", "ba6873"),
        ("melon", "ff7855"),
        ("sickly green", "94b21c"),
        ("silver", "c5c9c7"),
        ("purply blue", "661aee"),
        ("purpleish blue", "6140ef"),
        ("hospital green", "9be5aa"),
        ("shit brown", "7b5804"),
        ("mid blue", "276ab3"),
        ("amber", "feb308"),
        ("easter green", "8cfd7e"),
        ("soft blue", "6488ea"),
        ("cerulean blue", "056eee"),
        ("golden brown", "b27a01"),
        ("bright turquoise", "0ffef9"),
        ("red pink", "fa2a55"),
        ("red purple", "820747"),
        ("greyish brown", "7a6a4f"),
        ("vermillion", "f4320c"),
        ("russet", "a13905"),
        ("steel grey", "6f828a"),
        ("lighter purple", "a55af4"),
        ("bright violet", "ad0afd"),
        ("prussian blue", "004577"),
        ("slate green", "658d6d"),
        ("dirty pink", "ca7b80"),
        ("dark blue green", "005249"),
        ("pine", "2b5d34"),
        ("yellowy green", "bff128"),
        ("dark gold", "b59410"),
        ("bluish", "2976bb"),
        ("darkish blue", "014182"),
        ("dull red", "bb3f3f"),
        ("pinky red", "fc2647"),
        ("bronze", "a87900"),
        ("pale teal", "82cbb2"),
        ("military green", "667c3e"),
        ("barbie pink", "fe46a5"),
        ("bubblegum pink", "fe83cc"),
        ("pea soup green", "94a617"),
        ("dark mustard", "a88905"),
        ("shit", "7f5f00"),
        ("medium purple", "9e43a2"),
        ("very dark green", "062e03"),
        ("dirt", "8a6e45"),
        ("dusky pink", "cc7a8b"),
        ("red violet", "9e0168"),
        ("lemon yellow", "fdff38"),
        ("pistachio", "c0fa8b"),
        ("dull yellow", "eedc5b"),
        ("dark lime green", "7ebd01"),
        ("denim blue", "3b5b92"),
        ("teal blue", "01889f"),
        ("lightish blue", "3d7afd"),
        ("purpley blue", "5f34e7"),
        ("light indigo", "6d5acf"),
        ("swamp green", "748500"),
        ("brown green", "706c11"),
        ("dark maroon", "3c0008"),
        ("hot purple", "cb00f5"),
        ("dark forest green", "002d04"),
        ("faded blue", "658cbb"),
        ("drab green", "749551"),
        ("light lime green", "b9ff66"),
        ("snot green", "9dc100"),
        ("yellowish", "faee66"),
        ("light blue green", "7efbb3"),
        ("bordeaux", "7b002c"),
        ("light mauve", "c292a1"),
        ("ocean", "017b92"),
        ("marigold", "fcc006"),
        ("muddy green", "657432"),
        ("dull orange", "d8863b"),
        ("steel", "738595"),
        ("electric purple", "aa23ff"),
        ("fluorescent green", "08ff08"),
        ("yellowish brown", "9b7a01"),
        ("blush", "f29e8e"),
        ("soft green", "6fc276"),
        ("bright orange", "ff5b00"),
        ("lemon", "fdff52"),
        ("purple grey", "866f85"),
        ("acid green", "8ffe09"),
        ("pale lavender", "eecffe"),
        ("violet blue", "510ac9"),
        ("light forest green", "4f9153"),
        ("burnt red", "9f2305"),
        ("khaki green", "728639"),
        ("cerise", "de0c62"),
        ("faded purple", "916e99"),
        ("apricot", "ffb16d"),
        ("dark olive green", "3c4d03"),
        ("grey brown", "7f7053"),
        ("green grey", "77926f"),
        ("true blue", "010fcc"),
        ("pale violet", "ceaefa"),
        ("periwinkle blue", "8f99fb"),
        ("light sky blue", "c6fcff"),
        ("blurple", "5539cc"),
        ("green brown", "544e03"),
        ("bluegreen", "017a79"),
        ("bright teal", "01f9c6"),
        ("brownish yellow", "c9b003"),
        ("pea soup", "929901"),
        ("forest", "0b5509"),
        ("barney purple", "a00498"),
        ("ultramarine", "2000b1"),
        ("purplish", "94568c"),
        ("puke yellow", "c2be0e"),
        ("bluish grey", "748b97"),
        ("dark periwinkle", "665fd1"),
        ("dark lilac", "9c6da5"),
        ("reddish", "c44240"),
        ("light maroon", "a24857"),
        ("dusty purple", "825f87"),
        ("terra cotta", "c9643b"),
        ("avocado", "90b134"),
        ("marine blue", "01386a"),
        ("teal green", "25a36f"),
        ("slate grey", "59656d"),
        ("lighter green", "75fd63"),
        ("electric green", "21fc0d"),
        ("dusty blue", "5a86ad"),
        ("golden yellow", "fec615"),
        ("bright yellow", "fffd01"),
        ("light lavender", "dfc5fe"),
        ("umber", "b26400"),
        ("poop", "7f5e00"),
        ("dark peach", "de7e5d"),
        ("jungle green", "048243"),
        ("eggshell", "ffffd4"),
        ("denim", "3b638c"),
        ("yellow brown", "b79400"),
        ("dull purple", "84597e"),
        ("chocolate brown", "411900"),
        ("wine red", "7b0323"),
        ("neon blue", "04d9ff"),
        ("dirty green", "667e2c"),
        ("light tan", "fbeeac"),
        ("ice blue", "d7fffe"),
        ("cadet blue", "4e7496"),
        ("dark mauve", "874c62"),
        ("very light blue", "d5ffff"),
        ("grey purple", "826d8c"),
        ("pastel pink", "ffbacd"),
        ("very light green", "d1ffbd"),
        ("dark sky blue", "448ee4"),
        ("evergreen", "05472a"),
        ("dull pink", "d5869d"),
        ("aubergine", "3d0734"),
        ("mahogany", "4a0100"),
        ("reddish orange", "f8481c"),
        ("deep green", "02590f"),
        ("vomit green", "89a203"),
        ("purple pink", "e03fd8"),
        ("dusty pink", "d58a94"),
        ("faded green", "7bb274"),
        ("camo green", "526525"),
        ("pinky purple", "c94cbe"),
        ("pink purple", "db4bda"),
        ("brownish red", "9e3623"),
        ("dark rose", "b5485d"),
        ("mud", "735c12"),
        ("brownish", "9c6d57"),
        ("emerald green", "028f1e"),
        ("pale brown", "b1916e"),
        ("dull blue", "49759c"),
        ("burnt umber", "a0450e"),
        ("medium green", "39ad48"),
        ("clay", "b66a50"),
        ("light aqua", "8cffdb"),
        ("light olive green", "a4be5c"),
        ("brownish orange", "cb7723"),
        ("dark aqua", "05696b"),
        ("purplish pink", "ce5dae"),
        ("dark salmon", "c85a53"),
        ("greenish grey", "96ae8d"),
        ("jade", "1fa774"),
        ("ugly green", "7a9703"),
        ("dark beige", "ac9362"),
        ("emerald", "01a049"),
        ("pale red", "d9544d"),
        ("light magenta", "fa5ff7"),
        ("sky", "82cafc"),
        ("light cyan", "acfffc"),
        ("yellow orange", "fcb001"),
        ("reddish purple", "910951"),
        ("reddish pink", "fe2c54"),
        ("orchid", "c875c4"),
        ("dirty yellow", "cdc50a"),
        ("orange red", "fd411e"),
        ("deep red", "9a0200"),
        ("orange brown", "be6400"),
        ("cobalt blue", "030aa7"),
        ("neon pink", "fe019a"),
        ("rose pink", "f7879a"),
        ("greyish purple", "887191"),
        ("raspberry", "b00149"),
        ("aqua green", "12e193"),
        ("salmon pink", "fe7b7c"),
        ("tangerine", "ff9408"),
        ("brownish green", "6a6e09"),
        ("red brown", "8b2e16"),
        ("greenish brown", "696112"),
        ("pumpkin", "e17701"),
        ("pine green", "0a481e"),
        ("charcoal", "343837"),
        ("baby pink", "ffb7ce"),
        ("cornflower", "6a79f7"),
        ("blue violet", "5d06e9"),
        ("chocolate", "3d1c02"),
        ("greyish green", "82a67d"),
        ("scarlet", "be0119"),
        ("green yellow", "c9ff27"),
        ("dark olive", "373e02"),
        ("sienna", "a9561e"),
        ("pastel purple", "caa0ff"),
        ("terracotta", "ca6641"),
        ("aqua blue", "02d8e9"),
        ("sage green", "88b378"),
        ("blood red", "980002"),
        ("deep pink", "cb0162"),
        ("grass", "5cac2d"),
        ("moss", "769958"),
        ("pastel blue", "a2bffe"),
        ("bluish green", "10a674"),
        ("green blue", "06b48b"),
        ("dark tan", "af884a"),
        ("greenish blue", "0b8b87"),
        ("pale orange", "ffa756"),
        ("vomit", "a2a415"),
        ("forrest green", "154406"),
        ("dark lavender", "856798"),
        ("dark violet", "34013f"),
        ("purple blue", "632de9"),
        ("dark cyan", "0a888a"),
        ("olive drab", "6f7632"),
        ("pinkish", "d46a7e"),
        ("cobalt", "1e488f"),
        ("neon purple", "bc13fe"),
        ("light turquoise", "7ef4cc"),
        ("apple green", "76cd26"),
        ("dull green", "74a662"),
        ("wine", "80013f"),
        ("powder blue", "b1d1fc"),
        ("off white", "ffffe4"),
        ("electric blue", "0652ff"),
        ("dark turquoise", "045c5a"),
        ("blue purple", "5729ce"),
        ("azure", "069af3"),
        ("bright red", "ff000d"),
        ("pinkish red", "f10c45"),
        ("cornflower blue", "5170d7"),
        ("light olive", "acbf69"),
        ("grape", "6c3461"),
        ("greyish blue", "5e819d"),
        ("purplish blue", "601ef9"),
        ("yellowish green", "b0dd16"),
        ("greenish yellow", "cdfd02"),
        ("medium blue", "2c6fbb"),
        ("dusty rose", "c0737a"),
        ("light violet", "d6b4fc"),
        ("midnight blue", "020035"),
        ("bluish purple", "703be7"),
        ("red orange", "fd3c06"),
        ("dark magenta", "960056"),
        ("greenish", "40a368"),
        ("ocean blue", "03719c"),
        ("coral", "fc5a50"),
        ("cream", "ffffc2"),
        ("reddish brown", "7f2b0a"),
        ("burnt sienna", "b04e0f"),
        ("brick", "a03623"),
        ("sage", "87ae73"),
        ("grey green", "789b73"),
        ("white", "ffffff"),
        ("robin's egg blue", "98eff9"),
        ("moss green", "658b38"),
        ("steel blue", "5a7d9a"),
        ("eggplant", "380835"),
        ("light yellow", "fffe7a"),
        ("leaf green", "5ca904"),
        ("light grey", "d8dcd6"),
        ("puke", "a5a502"),
        ("pinkish purple", "d648d7"),
        ("sea blue", "047495"),
        ("pale purple", "b790d4"),
        ("slate blue", "5b7c99"),
        ("blue grey", "607c8e"),
        ("hunter green", "0b4008"),
        ("fuchsia", "ed0dd9"),
        ("crimson", "8c000f"),
        ("pale yellow", "ffff84"),
        ("ochre", "bf9005"),
        ("mustard yellow", "d2bd0a"),
        ("light red", "ff474c"),
        ("cerulean", "0485d1"),
        ("pale pink", "ffcfdc"),
        ("deep blue", "040273"),
        ("rust", "a83c09"),
        ("light teal", "90e4c1"),
        ("slate", "516572"),
        ("goldenrod", "fac205"),
        ("dark yellow", "d5b60a"),
        ("dark grey", "363737"),
        ("army green", "4b5d16"),
        ("grey blue", "6b8ba4"),
        ("seafoam", "80f9ad"),
        ("puce", "a57e52"),
        ("spring green", "a9f971"),
        ("dark orange", "c65102"),
        ("sand", "e2ca76"),
        ("pastel green", "b0ff9d"),
        ("mint", "9ffeb0"),
        ("light orange", "fdaa48"),
        ("bright pink", "fe01b1"),
        ("chartreuse", "c1f80a"),
        ("deep purple", "36013f"),
        ("dark brown", "341c02"),
        ("taupe", "b9a281"),
        ("pea green", "8eab12"),
        ("puke green", "9aae07"),
        ("kelly green", "02ab2e"),
        ("seafoam green", "7af9ab"),
        ("blue green", "137e6d"),
        ("khaki", "aaa662"),
        ("burgundy", "610023"),
        ("dark teal", "014d4e"),
        ("brick red", "8f1402"),
        ("royal purple", "4b006e"),
        ("plum", "580f41"),
        ("mint green", "8fff9f"),
        ("gold", "dbb40c"),
        ("baby blue", "a2cffe"),
        ("yellow green", "c0fb2d"),
        ("bright purple", "be03fd"),
        ("dark red", "840000"),
        ("pale blue", "d0fefe"),
        ("grass green", "3f9b0b"),
        ("navy", "01153e"),
        ("aquamarine", "04d8b2"),
        ("burnt orange", "c04e01"),
        ("neon green", "0cff0c"),
        ("bright blue", "0165fc"),
        ("rose", "cf6275"),
        ("light pink", "ffd1df"),
        ("mustard", "ceb301"),
        ("indigo", "380282"),
        ("lime", "aaff32"),
        ("sea green", "53fca1"),
        ("periwinkle", "8e82fe"),
        ("dark pink", "cb416b"),
        ("olive green", "677a04"),
        ("peach", "ffb07c"),
        ("pale green", "c7fdb5"),
        ("light brown", "ad8150"),
        ("hot pink", "ff028d"),
        ("black", "000000"),
        ("lilac", "cea2fd"),
        ("navy blue", "001146"),
        ("royal blue", "0504aa"),
        ("beige", "e6daa6"),
        ("salmon", "ff796c"),
        ("olive", "6e750e"),
        ("maroon", "650021"),
        ("bright green", "01ff07"),
        ("dark purple", "35063e"),
        ("mauve", "ae7181"),
        ("forest green", "06470c"),
        ("aqua", "13eac9"),
        ("cyan", "00ffff"),
        ("tan", "d1b26f"),
        ("dark blue", "00035b"),
        ("lavender", "c79fef"),
        ("turquoise", "06c2ac"),
        ("dark green", "033500"),
        ("violet", "9a0eea"),
        ("light purple", "bf77f6"),
        ("lime green", "89fe05"),
        ("grey", "929591"),
        ("sky blue", "75bbfd"),
        ("yellow", "ffff14"),
        ("magenta", "c20078"),
        ("light green", "96f97b"),
        ("orange", "f97306"),
        ("teal", "029386"),
        ("light blue", "95d0fc"),
        ("red", "e50000"),
        ("brown", "653700"),
        ("pink", "ff81c0"),
        ("blue", "0343df"),
        ("green", "15b01a"),
        ("purple", "7e1e9c"),
    ]
)

color_bytes = OrderedDict((key, bytes.fromhex(val)) for key, val in colors.items())



from __future__ import annotations

from functools import wraps
from typing import Any, Awaitable, Callable, NamedTuple, Optional, TypeVar, cast

from sanic.request import Request
from sanic.response import HTTPResponse, text

from .models import DataSourceId


class AccessRequest(NamedTuple):
    resourceId: DataSourceId
    resourceType: str
    mode: str

    @classmethod
    def read_dataset(
        cls, organization_name: str, dataset_name: str, **_: Any
    ) -> AccessRequest:
        return cls(
            resourceId=DataSourceId(organization_name, dataset_name),
            resourceType="datasource",
            mode="read",
        )

    @classmethod
    def administrate_datasets(cls, **_: Any) -> AccessRequest:
        return cls(
            resourceId=DataSourceId("", ""),
            resourceType="datasource",
            mode="administrate",
        )


class AccessAnswer(NamedTuple):
    granted: bool
    msg: Optional[str] = None


T = TypeVar("T", bound=Callable[..., Awaitable[HTTPResponse]])


def authorized(fn_access_request: Callable[..., AccessRequest]) -> Callable[[T], T]:
    def decorator(f: T) -> T:
        @wraps(f)
        async def decorated_function(
            request: Request, *args: Any, **kwargs: Any
        ) -> HTTPResponse:
            token = request.args.get("token", default="")
            parameters = request.app.router.get(
                request.path, request.method, request.host
            )[2]
            access_request = fn_access_request(**parameters)

            access_answer = await request.app.ctx.webknossos.request_access(
                token=token, access_request=access_request
            )

            if access_answer.granted:
                return await f(request, *args, **kwargs)
            else:
                return text(
                    f"Forbidden: {access_answer.msg}"
                    if access_answer.msg is not None
                    else "Token authentication failed",
                    status=403,
                )

        return cast(T, decorated_function)

    return decorator



from dataclasses import dataclass, replace
from typing import Dict, Optional, Tuple

from gcloud.aio.auth import Token

from ...utils.types import Box3D, Vec3D, Vec3Df
from ...webknossos.models import BoundingBox
from ...webknossos.models import DataLayer as WkDataLayer
from ...webknossos.models import DataSource as WkDataSource
from ...webknossos.models import DataSourceId as WkDataSourceId
from ..backend import DatasetInfo
from .meshes import MeshInfo
from .sharding import ShardingInfo


@dataclass(frozen=True)
class Scale:
    chunk_size: Vec3D
    encoding: str
    key: str
    mag: Vec3D
    size: Vec3D
    voxel_offset: Vec3D
    sharding: Optional[ShardingInfo]
    compressed_segmentation_block_size: Optional[Vec3D] = None

    def __post_init__(self) -> None:
        assert self.encoding in ["raw", "jpeg", "compressed_segmentation"]
        if self.encoding == "compressed_segmentation":
            assert self.compressed_segmentation_block_size is not None

    def box(self) -> Box3D:
        return Box3D.from_size(self.voxel_offset, self.size)

    def bounding_box(self) -> BoundingBox:
        return BoundingBox(self.voxel_offset, *self.size)

    def mag1_bounding_box(self) -> BoundingBox:
        return BoundingBox(self.voxel_offset * self.mag, *(self.size * self.mag))


@dataclass(frozen=True)
class Layer:
    source: str
    data_type: str
    num_channels: int
    scales: Tuple[Scale, ...]
    resolution: Vec3Df
    type: str
    mesh: Optional[MeshInfo] = None
    largestSegmentId: Optional[int] = None

    def __post_init__(self) -> None:
        supported_data_types = {
            "image": ["uint8", "uint16", "uint32", "uint64"],
            "segmentation": ["uint32", "uint64"],
        }
        assert self.type in supported_data_types
        assert self.data_type in supported_data_types[self.type]
        assert self.num_channels == 1
        min_mag = sorted([scale.mag for scale in self.scales])[0]
        assert all(scale.mag % min_mag == Vec3D.zeros() for scale in self.scales)

    def wk_data_type(self) -> str:
        return self.data_type

    def to_webknossos(self, layer_name: str) -> WkDataLayer:
        return WkDataLayer(
            layer_name,
            {"image": "color", "segmentation": "segmentation"}[self.type],
            self.scales[0].mag1_bounding_box(),
            [scale.mag for scale in self.scales],
            self.wk_data_type(),
            largestSegmentId=self.largestSegmentId,
        )


@dataclass
class Dataset(DatasetInfo):
    organization_name: str
    dataset_name: str
    layers: Dict[str, Layer]
    scale: Vec3Df
    token: Optional[Token] = None

    def to_webknossos(self) -> WkDataSource:
        return WkDataSource(
            WkDataSourceId(self.organization_name, self.dataset_name),
            [
                layer.to_webknossos(layer_name)
                for layer_name, layer in self.layers.items()
            ],
            self.scale,
        )

    @staticmethod
    def fix_scales(layers: Dict[str, Layer]) -> Tuple[Dict[str, Layer], Vec3Df]:
        """
        This function adapts the scale specification from neuroglancer to the
        webKnossos scheme. In neuroglancer, each `Scale` has an attached
        `resolution` (in nm), whereas webKnossos uses a single `scale` for all
        layers and differentiates through `Mag`s, which are power-of-two
        factors on the dataset's scale. In this function, the dataset's `scale`
        is determined and the `mag`s in the layers are computed and fixed.
        """
        layer_resolution_map = {
            layer_name: layer.resolution for layer_name, layer in layers.items()
        }
        min_resolution = sorted(list(layer_resolution_map.values()))[0]
        new_layers = {}
        for layer_name, layer in layers.items():
            assert layer.resolution % min_resolution == Vec3Df.zeros()
            factor_vec = layer.resolution / min_resolution
            new_scales = tuple(
                replace(scale, mag=(scale.mag.to_float() * factor_vec).to_int())
                for scale in layer.scales
            )
            new_layer = replace(layer, scales=new_scales, resolution=min_resolution)
            new_layers[layer_name] = new_layer
        return new_layers, min_resolution



import gzip
import logging
from pathlib import Path
from typing import Dict, List, Optional, cast

import h5py
import numpy as np
from aiohttp import ClientSession
from wkcuber.api.Dataset import WKDataset

from ...fast_wkw import DatasetCache  # pylint: disable=no-name-in-module
from ...utils.blocking import run_blocking
from ...utils.types import JSON, Vec3D
from ..backend import Backend, DatasetInfo
from .models import Dataset

logger = logging.getLogger()


class Wkw(Backend):
    wkw_cache: DatasetCache

    def __init__(self, config: Dict, http_client: ClientSession) -> None:
        super().__init__(config, http_client)
        cache_size = config.get("fileCacheMaxEntries", 1000)
        self.wkw_cache = DatasetCache(cache_size)

    async def handle_new_dataset(
        self, organization_name: str, dataset_name: str, dataset_info: JSON
    ) -> DatasetInfo:

        path = Wkw.path(dataset_info, organization_name, dataset_name)
        return Dataset(
            organization_name, dataset_name, WKDataset(str(path)), self.wkw_cache
        )

    async def read_data(
        self,
        abstract_dataset: DatasetInfo,
        layer_name: str,
        zoom_step: int,
        wk_offset: Vec3D,
        shape: Vec3D,
    ) -> Optional[np.ndarray]:
        dataset = cast(Dataset, abstract_dataset)
        return await dataset.read_data(layer_name, zoom_step, wk_offset, shape)

    async def get_meshes(
        self, abstract_dataset: DatasetInfo, layer_name: str
    ) -> Optional[List[str]]:
        dataset = cast(Dataset, abstract_dataset)
        assert (
            layer_name in dataset.dataset_handle.layers
        ), f"Layer {layer_name} does not exist"
        meshes_folder = dataset.dataset_handle.path / layer_name / "meshes"
        if meshes_folder.exists() and meshes_folder.is_dir:
            mesh_paths = list(meshes_folder.glob("*.hdf5"))
            return [mesh_path.name[:-5] for mesh_path in mesh_paths]
        return []

    async def get_chunks_for_mesh(
        self,
        abstract_dataset: DatasetInfo,
        layer_name: str,
        mesh_name: str,
        segment_id: int,
    ) -> Optional[List[Vec3D]]:
        dataset = cast(Dataset, abstract_dataset)
        assert (
            layer_name in dataset.dataset_handle.layers
        ), f"Layer {layer_name} does not exist"
        meshes_folder = dataset.dataset_handle.path / layer_name / "meshes"
        assert (
            meshes_folder.exists() and meshes_folder.is_dir()
        ), f"Mesh folder for layer {layer_name} does not exist"
        mesh_path = meshes_folder / f"{mesh_name}.hdf5"
        assert (
            meshes_folder.exists()
        ), f"Mesh {mesh_name} for layer {layer_name} does not exist"

        def read_mesh(mesh_path: Path, segment_id: int) -> List[Vec3D]:
            with h5py.File(mesh_path, "r") as mesh_file:
                segment_group = mesh_file.get(str(segment_id), None)
                if segment_group is None:
                    return []
                lod_group = segment_group["0"]
                chunk_keys = [
                    Vec3D(*[int(a) for a in chunk_key.split("_")])
                    for chunk_key in lod_group.keys()
                ]
                return chunk_keys

        return await run_blocking(read_mesh, mesh_path, segment_id)

    async def get_chunk_data_for_mesh(
        self,
        abstract_dataset: DatasetInfo,
        layer_name: str,
        mesh_name: str,
        segment_id: int,
        position: Vec3D,
    ) -> Optional[bytes]:
        dataset = cast(Dataset, abstract_dataset)
        assert (
            layer_name in dataset.dataset_handle.layers
        ), f"Layer {layer_name} does not exist"
        meshes_folder = dataset.dataset_handle.path / layer_name / "meshes"
        assert (
            meshes_folder.exists() and meshes_folder.is_dir()
        ), f"Mesh folder for layer {layer_name} does not exist"
        mesh_path = meshes_folder / f"{mesh_name}.hdf5"
        assert (
            meshes_folder.exists()
        ), f"Mesh {mesh_name} for layer {layer_name} does not exist"

        def read_chunk(mesh_path: Path, segment_id: int, position: Vec3D) -> bytes:
            with h5py.File(mesh_path, "r") as mesh_file:
                chunk_key = f"{position[0]}_{position[1]}_{position[2]}"
                mesh_encoding = mesh_file.attrs.get("metadata/encoding", "stl")
                mesh_binary = mesh_file[f"/{segment_id}/0/{chunk_key}"][:].tobytes()
                if mesh_encoding == "stl+gzip":
                    mesh_binary = gzip.decompress(mesh_binary)
                return mesh_binary

        return await run_blocking(read_chunk, mesh_path, segment_id, position)

    def clear_dataset_cache(self, abstract_dataset: DatasetInfo) -> None:
        dataset = cast(Dataset, abstract_dataset)
        dataset.clear_cache()

    @staticmethod
    def path(dataset_info: JSON, organization_name: str, dataset_name: str) -> Path:
        if "path" in dataset_info:
            return Path(dataset_info["path"])
        else:
            return Path("data", "binary", organization_name, dataset_name)



from sanic import Blueprint
from sanic.blueprint_group import BlueprintGroup

from .datasets import datasets
from .trace import trace
from .triggers import triggers

data = Blueprint.group(datasets, triggers, url_prefix="/data")
routes: BlueprintGroup = Blueprint.group(data, trace)



from abc import ABCMeta, abstractmethod
from typing import Dict, List, Optional

import numpy as np
from aiohttp import ClientSession

from ..utils.types import JSON, Vec3D
from ..webknossos.models import DataSource as WKDataSource


class DatasetInfo(metaclass=ABCMeta):
    organization_name: str
    dataset_name: str

    @abstractmethod
    def to_webknossos(self) -> WKDataSource:
        pass


class Backend(metaclass=ABCMeta):
    @classmethod
    def name(cls) -> str:
        return cls.__name__.lower()

    @abstractmethod
    def __init__(self, config: Dict, http_client: ClientSession) -> None:
        pass

    @abstractmethod
    async def handle_new_dataset(
        self, organization_name: str, dataset_name: str, dataset_info: JSON
    ) -> DatasetInfo:
        pass

    @abstractmethod
    async def read_data(
        self,
        dataset: DatasetInfo,
        layer_name: str,
        zoom_step: int,
        wk_offset: Vec3D,
        shape: Vec3D,
    ) -> Optional[np.ndarray]:
        """
        Read voxels from the backend.

        :param dataset:
        :param layer_name:
        :param zoomStep: 2^zoomStep is the smallest dimension of the scale
        :param wk_offset: in wk voxels
        :param shape: in scale voxels
        :returns: numpy array of shape shape
        """

    async def get_meshes(
        self, dataset: DatasetInfo, layer_name: str
    ) -> Optional[List[str]]:
        """
        List the available meshfiles of a layer from the backend.

        :param dataset:
        :param layer_name:
        :returns: list of meshfile names
        """
        raise NotImplementedError()

    async def get_chunks_for_mesh(
        self, dataset: DatasetInfo, layer_name: str, mesh_name: str, segment_id: int
    ) -> Optional[List[Vec3D]]:
        """
        List the available chunks of a mesh from the backend.

        :param dataset:
        :param layer_name:
        :param mesh_name:
        :param segment_id:
        :returns: list of tuples with the top-left position of each chunk
        """
        raise NotImplementedError()

    async def get_chunk_data_for_mesh(
        self,
        dataset: DatasetInfo,
        layer_name: str,
        mesh_name: str,
        segment_id: int,
        position: Vec3D,
    ) -> Optional[bytes]:
        """
        Read a chunk of a mesh from the backend.

        :param dataset:
        :param layer_name:
        :param mesh_name:
        :param segment_id:
        :param position:
        :returns: bytes of a mesh chunk in binary STL format
        """
        raise NotImplementedError()

    @abstractmethod
    def clear_dataset_cache(self, dataset: DatasetInfo) -> None:
        pass

    async def on_shutdown(self) -> None:
        pass



import traceback

from aiohttp import ClientResponseError


class RuntimeErrorWithUserMessage(RuntimeError):
    def __init__(self, user_message: str):
        self.user_message = user_message
        super().__init__(user_message)


def exception_traceback(exception: Exception) -> str:
    return "".join(
        traceback.format_exception(
            etype=type(exception), value=exception, tb=exception.__traceback__
        )
    )


def format_exception(exception: Exception) -> str:
    if isinstance(exception, ClientResponseError):
        description = f"HTTP {exception.message} ({exception.status})"
    if isinstance(exception, RuntimeErrorWithUserMessage):
        return exception.user_message
    else:
        description = type(exception).__name__
    return f"{description} in webknossos-connect"



from asyncio import get_running_loop
from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache
from typing import Any, Callable, TypeVar

THREAD_POOL_MAX_WORKERS = 32
_T = TypeVar("_T")


@lru_cache()
def get_thread_pool_executor() -> ThreadPoolExecutor:
    return ThreadPoolExecutor(max_workers=THREAD_POOL_MAX_WORKERS)


async def run_blocking(blocking_function: Callable[..., _T], *args: Any) -> _T:
    executor = get_thread_pool_executor()
    loop = get_running_loop()

    return await loop.run_in_executor(executor, blocking_function, *args)



import gzip
import math
from dataclasses import dataclass, field
from typing import Any, Optional, Tuple

import numpy as np

from ...utils.types import Vec3D
from .mmh3 import hash64


def compressed_morton_code(pos: Vec3D, grid_size: Vec3D) -> np.uint64:
    """
    Computes the compressed morton code as per
    https://github.com/google/neuroglancer/blob/master/src/neuroglancer/datasource/precomputed/volume.md#compressed-morton-code
    https://github.com/google/neuroglancer/blob/162b698f703c86e0b3e92b8d8e0cacb0d3b098df/src/neuroglancer/util/zorder.ts#L72
    """
    bits = Vec3D(
        math.ceil(math.log2(grid_size.x)),
        math.ceil(math.log2(grid_size.y)),
        math.ceil(math.log2(grid_size.z)),
    )

    max_bits = bits.as_np().max()
    output_bit = np.uint64(0)
    one = np.uint64(1)

    output = np.uint64(0)
    for bit in range(max_bits):
        if bit < bits.x:
            output |= ((np.uint64(pos.x) >> np.uint64(bit)) & one) << output_bit
            output_bit += one
        if bit < bits.y:
            output |= ((np.uint64(pos.y) >> np.uint64(bit)) & one) << output_bit
            output_bit += one
        if bit < bits.z:
            output |= ((np.uint64(pos.z) >> np.uint64(bit)) & one) << output_bit
            output_bit += one

    return output


def identity(a: np.uint64) -> np.uint64:
    return a


def compute_minishard_mask(minishard_bits: int) -> np.uint64:
    assert minishard_bits >= 0, "minishard_bits must be ≥ 0"
    if minishard_bits == 0:
        return np.uint64(0)

    minishard_mask = np.uint64(1)
    for i in range(minishard_bits - 1):
        minishard_mask <<= np.uint64(1)
        minishard_mask |= np.uint64(1)
    return np.uint64(minishard_mask)


def compute_shard_mask(shard_bits: int, minishard_bits: int) -> np.uint64:
    one_mask = np.uint64(0xFFFFFFFFFFFFFFFF)
    cursor = np.uint64(minishard_bits + shard_bits)
    shard_mask = ~((one_mask >> cursor) << cursor)
    minishard_mask = compute_minishard_mask(minishard_bits)
    return shard_mask & (~minishard_mask)  # pylint: disable=invalid-unary-operand-type


@dataclass(frozen=True)
class MinishardInfo:
    shard_number: np.uint64
    minishard_number: np.uint64


@dataclass(frozen=True)
class ShardingInfo:
    """
    Stores all necessary information for implementing the shared precomputed
    format of Neuroglancer. The actual downloading of the indices and chunk
    data needs to happen in the backend, because of the availability of the
    http client and controllable caching.
    """

    dataset_size: Vec3D
    chunk_size: Vec3D
    preshift_bits: int
    minishard_bits: int
    shard_bits: int
    hashfn: str
    minishard_index_encoding: str = "raw"
    data_encoding: str = "raw"
    _shard_mask: np.uint64 = field(init=False)
    _minishard_mask: np.uint64 = field(init=False)

    def __post_init__(self) -> None:
        # object.__setattr__ must be used to circumvent errors since the dataclass is frozen
        object.__setattr__(
            self,
            "_shard_mask",
            compute_shard_mask(self.shard_bits, self.minishard_bits),
        )
        object.__setattr__(
            self, "_minishard_mask", compute_minishard_mask(self.minishard_bits)
        )

    @staticmethod
    def parse(
        info_json: Any,
        dataset_size: Optional[Vec3D] = None,
        chunk_size: Optional[Vec3D] = None,
    ) -> "ShardingInfo":
        assert (
            info_json["@type"] == "neuroglancer_uint64_sharded_v1"
        ), "Only `neuroglancer_uint64_sharded_v1` sharding type is supported."

        return ShardingInfo(
            dataset_size=dataset_size if dataset_size is not None else Vec3D.zeros(),
            chunk_size=chunk_size if chunk_size is not None else Vec3D.zeros(),
            preshift_bits=int(info_json["preshift_bits"]),
            shard_bits=int(info_json["shard_bits"]),
            minishard_bits=int(info_json["minishard_bits"]),
            hashfn=info_json["hash"],
            minishard_index_encoding=info_json.get("minishard_index_encoding"),
            data_encoding=info_json.get("data_encoding"),
        )

    def get_chunk_key(self, pos: Vec3D) -> np.uint64:
        chunk_addr = pos // self.chunk_size
        grid_size = Vec3D(
            *np.ceil(np.array(self.dataset_size) / np.array(self.chunk_size))
        )
        key = compressed_morton_code(chunk_addr, grid_size)
        return key

    def format_shard_for_url(self, loc: MinishardInfo) -> str:
        return format(loc.shard_number, "x").zfill(int(np.ceil(self.shard_bits / 4.0)))

    def hash_chunk_id(self, chunk_id: np.uint64) -> np.uint64:
        assert self.hashfn in (
            "murmurhash3_x86_128",
            "identity",
        ), "Only `identity` or `murmurhash3_x86_128` hash functions are supported."
        if self.hashfn == "murmurhash3_x86_128":
            return np.uint64(hash64(chunk_id.tobytes())[0])
        else:
            return chunk_id

    def get_minishard_info(self, key: np.uint64) -> MinishardInfo:
        chunk_id = np.uint64(key) >> np.uint64(self.preshift_bits)
        chunk_id = self.hash_chunk_id(chunk_id)
        minishard_number = np.uint64(chunk_id & self._minishard_mask)
        shard_number = np.uint64(
            (chunk_id & self._shard_mask) >> np.uint64(self.minishard_bits)
        )
        return MinishardInfo(shard_number, minishard_number)

    def get_shard_index_range(self) -> Tuple[int, int]:
        shard_index_end = (np.uint64(1) << np.uint64(self.minishard_bits)) * 16
        return (0, shard_index_end)

    def parse_shard_index(self, buf: bytes) -> np.ndarray:
        # See: https://github.com/google/neuroglancer/blob/master/src/neuroglancer/datasource/precomputed/sharded.md#shard-index-format
        # Format: [[minishard_index_start, minishard_index_end], ...]
        # Offsets are relative to shard_index_end
        return (
            np.frombuffer(buf, dtype="<u8")
            .reshape((np.uint64(1) << np.uint64(self.minishard_bits), 2))
            .copy()
        )

    def get_minishard_index_range(
        self, minishard_number: np.uint64, shard_index: np.ndarray
    ) -> Tuple[int, int]:
        _, shard_index_end = self.get_shard_index_range()
        minishard_index_start = shard_index_end + shard_index[minishard_number, 0]
        minishard_index_end = shard_index_end + shard_index[minishard_number, 1]
        return (minishard_index_start, minishard_index_end)

    def parse_minishard_index(self, buf: bytes) -> np.ndarray:
        if self.minishard_index_encoding == "gzip":
            buf = gzip.decompress(buf)
        # See: https://github.com/google/neuroglancer/blob/master/src/neuroglancer/datasource/precomputed/sharded.md#minishard-index-format
        # Format: [[chunk_id, chunk_offset, chunk_byte_count], ...]
        # Offsets are relative to shard_index_end
        index = np.frombuffer(buf, dtype="<u8").reshape((3, -1)).copy()
        if index.shape[1] == 0:
            return index.T
        index[0, :] = np.cumsum(index[0])
        index[1, 1:] = index[1, 0] + np.cumsum(index[1, 1:]) + np.cumsum(index[2, 0:-1])
        return index.T

    def get_chunk_range(
        self, chunk_id: np.uint64, minishard_index: np.ndarray
    ) -> Optional[Tuple[int, int]]:
        _, shard_index_end = self.get_shard_index_range()
        chunk_entries = minishard_index[minishard_index[:, 0] == chunk_id]
        if chunk_entries.shape[0] == 0:
            return None
        chunk_entry = chunk_entries[0]
        return (
            int(shard_index_end + chunk_entry[1]),
            int(shard_index_end + chunk_entry[2] + chunk_entry[1]),
        )

    def parse_chunk(self, buf: bytes) -> bytes:
        if self.data_encoding == "gzip":
            return gzip.decompress(buf)
        return buf



import asyncio
import json
import os
import tempfile
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, cast

import compressed_segmentation
import jpeg4py as jpeg
import numpy as np
from aiohttp import ClientResponseError, ClientSession
from async_lru import alru_cache
from gcloud.aio.auth import Token

from ...utils.exceptions import RuntimeErrorWithUserMessage
from ...utils.types import JSON, Box3D, Vec3D, Vec3Df
from ..backend import Backend, DatasetInfo
from ..neuroglancer.meshes import Meshfile, MeshfileLod, MeshInfo
from .models import Dataset, Layer, Scale
from .sharding import MinishardInfo, ShardingInfo

DecoderFn = Callable[[bytes, str, Vec3D, Optional[Vec3D]], np.ndarray]
Chunk = Tuple[Box3D, np.ndarray]

MESH_LOD = 2
MESH_NAME = "mesh"


class NeuroglancerAuthenticationError(RuntimeError):
    pass


class NeuroglancerAuthenticationMissingError(RuntimeErrorWithUserMessage):
    pass


async def get_header(token: Optional[Token]) -> Dict[str, str]:
    if token is None:
        return {}
    else:
        try:
            token_str = await token.get()
        except Exception as e:
            raise NeuroglancerAuthenticationError(*e.args)
        if token_str is None:
            return {}
        return {"Authorization": "Bearer " + token_str}


def select_lod(meshfile: Meshfile) -> MeshfileLod:
    return meshfile.lods[min(len(meshfile.lods) - 1, MESH_LOD)]


class Neuroglancer(Backend):
    def __init__(self, config: Dict, http_client: ClientSession) -> None:
        self.http_client = http_client
        self.decoders: Dict[str, DecoderFn] = {
            "raw": self.__decode_raw,
            "jpeg": self.__decode_jpeg,
            "compressed_segmentation": self.__decode_compressed_segmentation,
        }

    def create_token(self, credentials: Optional[JSON]) -> Optional[Token]:
        if credentials is None:
            return None
        tmpfile = tempfile.NamedTemporaryFile(mode="w+", delete=False)
        json.dump(credentials, tmpfile)
        tmpfile.close()
        try:
            token = Token(
                service_file=tmpfile.name,
                session=self.http_client,
                scopes=["https://www.googleapis.com/auth/cloud-platform"],
            )
        finally:
            os.remove(tmpfile.name)
        return token

    async def __handle_layer(
        self, layer_name: str, layer: Dict[str, Any], token: Optional[Token]
    ) -> Tuple[str, Layer]:
        layer["source"] = (
            layer["source"]
            .replace("gs://", "https://storage.googleapis.com/")
            .replace("s3://", "https://s3.amazonaws.com/")
        )
        info_url = layer["source"] + "/info"

        try:
            async with await self.http_client.get(
                info_url, headers=await get_header(token)
            ) as r:
                response = await r.json(content_type=None)
        except ClientResponseError as e:
            if e.status == 403:
                raise NeuroglancerAuthenticationMissingError(
                    "Could not authenticate the neuroglancer dataset, "
                    + "please add authentication"
                )
            else:
                raise e
        layer.update(response)

        if "mesh" in layer:
            mesh_info_url = layer["source"] + "/" + layer["mesh"] + "/info"
            try:
                async with await self.http_client.get(
                    mesh_info_url, headers=await get_header(token)
                ) as r:
                    mesh_info = await r.json(content_type=None)
                    layer["mesh"] = MeshInfo.parse(mesh_info)
            except ClientResponseError as e:
                if e.status == 403:
                    raise NeuroglancerAuthenticationMissingError(
                        "Could not authenticate the neuroglancer dataset, "
                        + "please add authentication"
                    )
                elif e.status == 404:
                    del layer["mesh"]
                else:
                    raise e

        layer["scales"] = sorted(layer["scales"], key=lambda scale: scale["resolution"])
        min_resolution = Vec3Df(*layer["scales"][0]["resolution"])

        new_scales: List[Scale] = []
        for scale in layer["scales"]:
            resolution = Vec3Df(*scale["resolution"])
            chunk_size = Vec3D(*scale["chunk_sizes"][0])
            dataset_size = Vec3D(*scale["size"])
            assert resolution % min_resolution == Vec3Df.zeros()
            new_scales.append(
                Scale(
                    chunk_size=chunk_size,
                    encoding=scale["encoding"],
                    key=scale["key"],
                    mag=(resolution // min_resolution).to_int(),
                    size=dataset_size,
                    voxel_offset=(
                        Vec3D(*scale["voxel_offset"]).to_int()
                        if "voxel_offset" in scale
                        else Vec3D.zeros()
                    ),
                    sharding=ShardingInfo.parse(
                        scale["sharding"], dataset_size, chunk_size
                    )
                    if "sharding" in scale
                    else None,
                    compressed_segmentation_block_size=Vec3D(
                        *scale["compressed_segmentation_block_size"]
                    )
                    if "compressed_segmentation_block_size" in scale
                    else None,
                )
            )

        return (
            layer_name,
            Layer(
                source=layer["source"],
                data_type=layer["data_type"],
                num_channels=int(layer["num_channels"]),
                scales=tuple(new_scales),
                resolution=min_resolution,
                type=layer["type"],
                mesh=layer.get("mesh"),
            ),
        )

    async def handle_new_dataset(
        self, organization_name: str, dataset_name: str, dataset_info: JSON
    ) -> DatasetInfo:
        try:
            token = self.create_token(dataset_info.get("credentials", None))
        except Exception as e:
            NeuroglancerAuthenticationError(*e.args)
        layers = dict(
            await asyncio.gather(
                *(
                    self.__handle_layer(layer_name, layer, token)
                    for layer_name, layer in dataset_info["layers"].items()
                )
            )
        )
        layers, scale = Dataset.fix_scales(layers)
        dataset = Dataset(organization_name, dataset_name, layers, scale, token)
        return dataset

    def __decode_raw(
        self, buffer: bytes, data_type: str, chunk_size: Vec3D, _: Optional[Vec3D]
    ) -> np.ndarray:
        return np.frombuffer(buffer, dtype=data_type).reshape(chunk_size, order="F")

    def __decode_jpeg(
        self, buffer: bytes, data_type: str, chunk_size: Vec3D, _: Optional[Vec3D]
    ) -> np.ndarray:
        np_bytes = np.frombuffer(buffer, dtype="uint8")
        return (
            jpeg.JPEG(np_bytes)
            .decode()[:, :, 0]
            .astype(data_type)
            .T.reshape(chunk_size, order="F")
        )

    def __decode_compressed_segmentation(
        self,
        buffer: bytes,
        data_type: str,
        chunk_size: Vec3D,
        block_size: Optional[Vec3D],
    ) -> np.ndarray:
        assert block_size is not None
        return compressed_segmentation.decompress(
            buffer,
            chunk_size.as_tuple(),
            dtype=data_type,
            block_size=block_size,
            order="F",
        )

    def __chunks(
        self, requested: Box3D, ds_frame: Box3D, wk_chunk_size: Vec3D
    ) -> Iterable[Box3D]:
        # clip data outside available data
        inside = requested.intersect(ds_frame)

        # align request to chunks
        aligned = (inside - ds_frame.left).div(
            wk_chunk_size
        ) * wk_chunk_size + ds_frame.left

        for chunk_offset in aligned.range(offset=wk_chunk_size):
            chunk = Box3D.from_size(chunk_offset, wk_chunk_size)
            # chunk is capped to fit the dataset:
            yield chunk.intersect(ds_frame)

    async def __read_ranged_data(
        self, url: str, range: Tuple[int, int], token: Optional[Token]
    ) -> bytes:
        if range[0] == range[1]:
            return b""
        headers = {
            **await get_header(token),
            "Range": f"bytes={int(range[0])}-{int(range[1]-1)}",
        }
        async with await self.http_client.get(url, headers=headers) as r:
            r.raise_for_status()
            response_buffer = await r.read()
        return response_buffer

    @alru_cache(maxsize=2 ** 12, cache_exceptions=False)
    async def __read_shard_index(
        self, shard_url: str, sharding_info: ShardingInfo, token: Optional[Token]
    ) -> np.ndarray:
        shard_index_range = sharding_info.get_shard_index_range()
        index = sharding_info.parse_shard_index(
            await self.__read_ranged_data(shard_url, shard_index_range, token)
        )
        return index

    @alru_cache(maxsize=2 ** 12, cache_exceptions=False)
    async def __read_minishard_index(
        self,
        shard_url: str,
        sharding_info: ShardingInfo,
        minishard_info: MinishardInfo,
        token: Optional[Token],
    ) -> np.ndarray:
        shard_index = await self.__read_shard_index(shard_url, sharding_info, token)
        minishard_index_range = sharding_info.get_minishard_index_range(
            minishard_info.minishard_number, shard_index
        )
        buf = await self.__read_ranged_data(shard_url, minishard_index_range, token)
        index = sharding_info.parse_minishard_index(buf)
        return index

    async def __read_sharded_chunk(
        self,
        source_url: str,
        sharding_info: ShardingInfo,
        position: Vec3D,
        token: Optional[Token],
    ) -> Optional[bytes]:
        chunk_id = sharding_info.get_chunk_key(position)
        minishard_info = sharding_info.get_minishard_info(chunk_id)
        shard_url = (
            f"{source_url}/{sharding_info.format_shard_for_url(minishard_info)}.shard"
        )
        minishard_index = await self.__read_minishard_index(
            shard_url, sharding_info, minishard_info, token
        )
        chunk_range = sharding_info.get_chunk_range(chunk_id, minishard_index)
        if chunk_range is None:
            return None
        buf = sharding_info.parse_chunk(
            await self.__read_ranged_data(shard_url, chunk_range, token)
        )

        return buf

    @alru_cache(maxsize=2 ** 12, cache_exceptions=False)
    async def __read_chunk(
        self,
        layer: Layer,
        scale: Scale,
        chunk_box: Box3D,
        decoder_fn: DecoderFn,
        token: Optional[Token],
    ) -> Chunk:
        response_buffer = None
        if scale.sharding is not None:
            response_buffer = await self.__read_sharded_chunk(
                f"{layer.source}/{scale.key}", scale.sharding, chunk_box.left, token
            )
            if response_buffer is None:
                return (
                    chunk_box,
                    np.zeros(chunk_box.size().as_tuple(), dtype=layer.wk_data_type()),
                )
        else:
            url_coords = "_".join(
                f"{left_dim}-{right_dim}" for left_dim, right_dim in zip(*chunk_box)
            )
            data_url = f"{layer.source}/{scale.key}/{url_coords}"

            async with await self.http_client.get(
                data_url, headers=await get_header(token)
            ) as r:
                response_buffer = await r.read()

        chunk_data = decoder_fn(
            response_buffer,
            layer.data_type,
            chunk_box.size(),
            scale.compressed_segmentation_block_size,
        ).astype(layer.wk_data_type())
        return (chunk_box, chunk_data)

    def __cutout(self, chunks: List[Chunk], box: Box3D) -> np.ndarray:
        result = np.zeros(box.size(), dtype=chunks[0][1].dtype, order="F")
        for chunk_box, chunk_data in chunks:
            inner = chunk_box.intersect(box)
            result[(inner - box.left).np_slice()] = chunk_data[
                (inner - chunk_box.left).np_slice()
            ]
        return result

    async def read_data(
        self,
        abstract_dataset: DatasetInfo,
        layer_name: str,
        zoom_step: int,
        wk_offset: Vec3D,
        shape: Vec3D,
    ) -> Optional[np.ndarray]:
        dataset = cast(Dataset, abstract_dataset)
        layer = dataset.layers[layer_name]
        # we can use zoom_step as the index, as the scales are sorted
        # see assertion in the Layer initialization
        scale = [
            scale for scale in layer.scales if 2 ** zoom_step == scale.mag.as_np().max()
        ][0]
        decoder = self.decoders[scale.encoding]

        # convert to coordinate system of current scale
        offset = wk_offset // scale.mag
        box = Box3D.from_size(offset, shape)

        chunk_boxes = self.__chunks(box, scale.box(), scale.chunk_size)
        try:
            chunks = await asyncio.gather(
                *(
                    self.__read_chunk(layer, scale, chunk_box, decoder, dataset.token)
                    for chunk_box in chunk_boxes
                )
            )
        except ClientResponseError:
            # will be reported in MISSING-BUCKETS, frontend will retry
            return None

        return self.__cutout(chunks, box)

    @alru_cache(maxsize=2 ** 12, cache_exceptions=False)
    async def __read_meshfile(
        self, layer: Layer, segment_id: int, token: Optional[Token]
    ) -> Optional[Meshfile]:
        assert layer.mesh is not None
        mesh_info = layer.mesh
        sharding_info = mesh_info.sharding
        chunk_id = np.uint64(segment_id)
        minishard_info = sharding_info.get_minishard_info(chunk_id)
        shard_url = f"{layer.source}/mesh/{sharding_info.format_shard_for_url(minishard_info)}.shard"
        minishard_index = await self.__read_minishard_index(
            shard_url, sharding_info, minishard_info, token
        )
        chunk_range = sharding_info.get_chunk_range(chunk_id, minishard_index)
        # print("chunk_range", chunk_range)
        if chunk_range is None:
            return None
        buf = sharding_info.parse_chunk(
            await self.__read_ranged_data(shard_url, chunk_range, token)
        )
        return mesh_info.parse_meshfile(buf, chunk_range[0])

    async def __read_mesh_data(
        self,
        layer: Layer,
        meshfile: Meshfile,
        segment_id: int,
        position: Vec3D,
        token: Optional[Token],
    ) -> Optional[bytes]:
        assert layer.mesh is not None
        sharding_info = layer.mesh.sharding
        fragment = next(
            fragment
            for fragment in select_lod(meshfile).fragments
            if fragment.position.to_int() == position
        )
        chunk_id = np.uint64(segment_id)
        minishard_info = sharding_info.get_minishard_info(chunk_id)
        shard_url = f"{layer.source}/mesh/{sharding_info.format_shard_for_url(minishard_info)}.shard"
        buf = await self.__read_ranged_data(
            shard_url,
            (fragment.byte_offset, fragment.byte_offset + fragment.byte_size),
            token,
        )
        return meshfile.decode_data(fragment, buf)

    async def get_meshes(
        self, abstract_dataset: DatasetInfo, layer_name: str
    ) -> Optional[List[str]]:
        dataset = cast(Dataset, abstract_dataset)
        if layer_name in dataset.layers and dataset.layers[layer_name].mesh:
            return [MESH_NAME]
        return []

    async def get_chunks_for_mesh(
        self,
        abstract_dataset: DatasetInfo,
        layer_name: str,
        mesh_name: str,
        segment_id: int,
    ) -> Optional[List[Vec3D]]:
        dataset = cast(Dataset, abstract_dataset)
        layer = dataset.layers[layer_name]
        assert layer.mesh is not None
        assert mesh_name == MESH_NAME
        meshfile = await self.__read_meshfile(layer, segment_id, dataset.token)
        if meshfile is None:
            return None
        return [
            fragment.position.to_int() for fragment in select_lod(meshfile).fragments
        ]

    async def get_chunk_data_for_mesh(
        self,
        abstract_dataset: DatasetInfo,
        layer_name: str,
        mesh_name: str,
        segment_id: int,
        position: Vec3D,
    ) -> Optional[bytes]:
        dataset = cast(Dataset, abstract_dataset)
        layer = dataset.layers[layer_name]
        assert layer.mesh is not None
        assert mesh_name == MESH_NAME
        meshfile = await self.__read_meshfile(layer, segment_id, dataset.token)
        if meshfile is None:
            return None
        return await self.__read_mesh_data(
            layer, meshfile, segment_id, position, dataset.token
        )

    def clear_dataset_cache(self, dataset: DatasetInfo) -> None:
        self.__read_chunk.cache_clear()  # pylint: disable=no-member
        self.__read_shard_index.cache_clear()  # pylint: disable=no-member
        self.__read_minishard_index.cache_clear()  # pylint: disable=no-member
        self.__read_meshfile.cache_clear()  # pylint: disable=no-member
